unit CreateReq;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, wcrypt2, Buttons;

type
  TCreateReqForm = class(TForm)
    Label2: TLabel;
    ContainerEdit: TEdit;
    SubjectEdit: TEdit;
    OKBtn: TBitBtn;
    SaveDlg: TSaveDialog;
    procedure OKBtnClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  CreateReqForm: TCreateReqForm;

implementation

{$R *.DFM}

procedure TCreateReqForm.OKBtnClick(Sender: TObject);
var nameAttr: CERT_RDN_ATTR;
    nameString: PChar;
    rdn: CERT_RDN;
    nameInfo: CERT_NAME_INFO;
    certReqInfo: CERT_REQUEST_INFO;
    subjNameBlob: CERT_NAME_BLOB;
    encNameLen: DWORD;
    encName: PBYTE;
    prov: HCRYPTPROV;
    pubKeyInfoLen: DWORD;
    pubKeyInfo: PCERT_PUBLIC_KEY_INFO;
    encCertReqLen: DWORD;
    params: CRYPT_OBJID_BLOB;
    sigAlg: CRYPT_ALGORITHM_IDENTIFIER;
    signedEncCertReq: PBYTE;
    cont: PChar;
    err: string;
    KeyExchKey, SignKey: HCRYPTKEY;
    encType: DWORD;
    f: file;

begin

//1.    CERT_RDN_ATTR       :
//      ,
//      ,
//     ,
//       ,    PBYTE.

//   RDN  Relative Distinguished Name (  ).
//   RDN  , ,  , , ,     ..
//        :
//      ,      - "2.5.4.3",
//         - "2.5.4.6", ..
//          - ,
//      CERT_RDN_PRINTABLE_STRING  CERT_RDN_ENCODED_BLOB.

nameString := StrAlloc(length(SubjectEdit.text)+1);
StrPCopy(nameString, SubjectEdit.Text);
encType := PKCS_7_ASN_ENCODING or X509_ASN_ENCODING;
nameAttr.pszObjId := '2.5.4.3';
nameAttr.dwValueType := CERT_RDN_PRINTABLE_STRING;
nameAttr.Value.cbData := length(SubjectEdit.text);
nameAttr.Value.pbData := PBYTE(nameString);


//  2.    CERT_RDN       
//     CERT_RDN_ATTR,  ,       .
rdn.cRDNAttr := 1;
rdn.rgRDNAttr := @nameAttr;

//  3.  CERT_NAME_INFO        ,
//     3,       .
nameInfo.cRDN := 1;
nameInfo.rgRDN := @rdn;

//  4.   CryptEncodeObject,      
//   CERT_NAME_INFO.       CERT_NAME_BLOB,
//       .
//         ;   CryptoAPI 
//        : PKCS_7_ASN_ENCODING or X509_ASN_ENCODING
if not CryptEncodeObject(encType, X509_NAME, @nameInfo, nil, @encNameLen) or (encNameLen < 1) then
   begin
   MessageDlg('CryptEncodeObject error ' + inttostr(GetLastError), mtError, [mbOK], 0);
   StrDispose(nameString);
   exit;
   end;

GetMem(encName, encNameLen);

if not CryptEncodeObject(PKCS_7_ASN_ENCODING or X509_ASN_ENCODING, X509_NAME,
       @nameInfo, encName, @encNameLen) or (encNameLen < 1) then
   begin
   MessageDlg('2nd CryptEncodeObject error ' + inttostr(GetLastError), mtError, [mbOK], 0);
   StrDispose(nameString);
   FreeMem(encName, encNameLen);
   exit;
   end;


//  5.   Subject  CERT_REQUEST_INFO    CERT_NANE_BLOB

//  6.         ,
//          CRYPT_ATTR_BLOB,
//         CRYPT_ATTRIBUTE.
//              
//   CERT_REQUEST_INFO.

subjNameBlob.cbData := encNameLen;
subjNameBlob.pbData := encName;


// 7.  CERT_REQUEST_INFO     .
// CryptoAPI    :
//    1 -     ;
//    2 -        ;
//    3 -       ( ),
//          .

certReqInfo.Subject := subjNameBlob;
certReqInfo.cAttribute := 0;
certReqInfo.rgAttribute := nil;
certReqInfo.dwVersion := CERT_REQUEST_V1;

if length(ContainerEdit.Text) = 0
then cont := nil
else
    begin
    err := ContainerEdit.Text;
    cont := StrAlloc(length(err) + 1);
    StrPCopy(cont, err);
    end;

if not CryptAcquireContext(@prov, cont, MS_DEF_PROV, PROV_RSA_FULL, 0)
then
    begin
    MessageDlg('  : ' + inttostr(GetLastError), mtError, [mbOK], 0);
    StrDispose(nameString);
    FreeMem(encName, encNameLen);
    if cont <> nil then StrDispose(cont);
    exit;
    end;


{if not CryptGenKey(prov, AT_KEYEXCHANGE, 0, @KeyExchKey)
then
    begin
        MessageDlg('    : ',  mtError, [mbOK], 0)
    end;
}

//  8.  CryptExportPublicKeyInfo,    
//  CERT_PUBLIC_KEY_INFO,     .
if not CryptExportPublicKeyInfo (prov, AT_SIGNATURE, encType, nil, @pubKeyInfoLen) then
   begin
   MessageDlg('   ', mtError, [mbOK], 0);
   StrDispose(nameString);
   FreeMem(encName, encNameLen);
   if cont <> nil then StrDispose(cont);
   exit;
   end;

GetMem(pubKeyInfo, pubKeyInfoLen);

if not CryptExportPublicKeyInfo (prov, AT_SIGNATURE, encType, pubKeyInfo, @pubKeyInfoLen) then
   begin
   MessageDlg('   ', mtError, [mbOK], 0);
   StrDispose(nameString);
   FreeMem(encName, encNameLen);
   if cont <> nil then StrDispose(cont);
   FreeMem(pubKeyInfo, pubKeyInfoLen);
   exit;
   end;


//  9.   SubjectPublicKeyInfo  CERT_REQUEST_INFO   .
certReqInfo.SubjectPublicKeyInfo := pubKeyInfo^;
FillChar(params, sizeof(params), 0);
sigAlg.pszObjId := szOID_OIWSEC_sha1RSASign;
sigAlg.Parameters := params;


//  10.  CryptSignAndEncodeCertificate,       CERT_REQUEST_INFO.
//      CERT_REQUEST_INFO   ,
//        ,     
//       .

if not CryptSignAndEncodeCertificate (prov, AT_SIGNATURE, encType,
   X509_CERT_REQUEST_TO_BE_SIGNED, @certReqInfo, @sigAlg, nil, nil,
   @encCertReqLen) then
   begin
   MessageDlg('    ', mtError, [mbOK], 0);
   StrDispose(nameString);
   FreeMem(encName, encNameLen);
   if cont <> nil then StrDispose(cont);
   FreeMem(pubKeyInfo, pubKeyInfoLen);
   exit;
   end;

GetMem(signedEncCertReq, encCertReqLen);

if not CryptSignAndEncodeCertificate (prov, AT_SIGNATURE, encType, X509_CERT_REQUEST_TO_BE_SIGNED,
   @certReqInfo, @sigAlg, nil, signedEncCertReq, @encCertReqLen) then
   begin
   MessageDlg('   ', mtError, [mbOK], 0);
   StrDispose(nameString);
   FreeMem(encName, encNameLen);
   if cont <> nil then StrDispose(cont);
   FreeMem(pubKeyInfo, pubKeyInfoLen);
   FreeMem(signedEncCertReq, encCertReqLen);
   exit;
   end;

SaveDlg.Title := '   ?';
SaveDlg.Filter := '  (*.req)|*.req|  (*.*)|*.*';
SaveDlg.DefaultExt := 'req';

if SaveDlg.Execute then
   begin
   AssignFile(f, SaveDlg.FileName);
   rewrite(f, 1);
   BlockWrite(f, signedEncCertReq^, encCertReqLen);
   CloseFile(f);
   end;

StrDispose(nameString);
FreeMem(encName, encNameLen);

if cont <> nil then StrDispose(cont);

FreeMem(pubKeyInfo, pubKeyInfoLen);
FreeMem(signedEncCertReq, encCertReqLen);

if not CryptReleaseContext(prov, 0) then
   begin
   MessageDlg('  ', mtError, [mbOK], 0);
   end;
createreqform.Close;
end;
end.

