unit SignCert;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, wcrypt2, ExtCtrls;

type
  TSignCertForm = class(TForm)
    OpenDlg: TOpenDialog;
    SaveDlg: TSaveDialog;
    OpenReqButton: TButton;
    Label1: TLabel;
    ContainerEdit: TEdit;
    SubjectEdit: TEdit;
    SignCheckBox: TCheckBox;
    Label4: TLabel;
    SerNumEdit: TEdit;
    Label5: TLabel;
    NotBeforeEdit: TEdit;
    Label6: TLabel;
    NotAfterEdit: TEdit;
    Label7: TLabel;
    IssuerEdit: TEdit;
    Splitter1: TSplitter;
    SignButton: TButton;
    procedure OpenReqButtonClick(Sender: TObject);
    procedure SignButtonClick(Sender: TObject);
  private
    pCertReqInfo: PCERT_REQUEST_INFO;
  public

  end;

var
  SignCertForm: TSignCertForm;

implementation

{$R *.DFM}

procedure TSignCertForm.OpenReqButtonClick(Sender: TObject);
var f: file;
    reqEncoded, buf, buf2: PBYTE;
    size, rsize: DWORD;
    encType: DWORD;
    subjNameBLOB: CERT_NAME_BLOB;
    subjNameString, cont: PChar;
    signedContent: PCERT_SIGNED_CONTENT_INFO;
    err: string;
    p: pointer;
    DERBLOB: CRYPT_DER_BLOB;
    prov: HCRYPTPROV;
begin

if length(ContainerEdit.Text) = 0
then cont := nil
else
    begin
    err := ContainerEdit.Text;
    cont := StrAlloc(length(err) + 1);
    StrPCopy(cont, err);
    end;

if not CryptAcquireContext(@prov, cont, MS_DEF_PROV, PROV_RSA_FULL, 0)
then
    begin
      MessageDlg('  : ', mtError, [mbOK], 0);
      exit;
    end;

//       
//      ( reqEncoded).
subjNameString := StrAlloc(512);
encType := PKCS_7_ASN_ENCODING or X509_ASN_ENCODING;

OpenDlg.Title := '   ?';
OpenDlg.Filter := '  (*.req)|*.req|  (*.*)|*.*';
OpenDlg.DefaultExt := 'req';
if OpenDlg.Execute then
   begin
   assignFile(f, OpenDlg.FileName);
   reset(f, 1);
   size := FileSize(f);
   GetMem(reqEncoded, size);
   BlockRead(f, reqEncoded^, size);
   CloseFile(f);
   end
else exit;


// 
GetMem(buf, 2048);
rsize := 2048;
if not CryptDecodeObject(encType, X509_CERT, reqEncoded, size, 0, buf, @rsize) then
   begin
   case int64(GetLastError) of
   ERROR_FILE_NOT_FOUND: err := 'File not found';
   ERROR_MORE_DATA: err := 'More data';
   else err := 'some error';
   end;
   MessageDlg('Error decoding request ' + err, mtError, [mbOK], 0);
   FreeMem(reqEncoded, size);
   FreeMem(buf, 2048);
   FreeMem(subjNameString, 512);
   exit;
   end;

//       ,    -
//  CERT_REQUEST_INFO
p := buf;
signedContent := p;
DERBLOB := signedContent^.toBeSigned;
rsize := 512;
GetMem(buf2, rsize);
if not CryptDecodeObject(encType, X509_CERT_REQUEST_TO_BE_SIGNED,
   DERBLOB.pbData, DERBLOB.cbData, 0, buf2, @rsize) then
   begin
   case int64(GetLastError) of
   ERROR_FILE_NOT_FOUND: err := 'File not found';
   ERROR_MORE_DATA: err := 'More data';
   else err := 'some error';
   end;
   MessageDlg('Error decoding DERBLOB ' + err, mtError, [mbOK], 0);
   FreeMem(reqEncoded, size);
   FreeMem(buf, 2048);
   FreeMem(subjNameString, 512);
   FreeMem(buf2, 512);
   exit;
   end;

p := buf2;
pCertReqInfo := p;
buf2 := nil;

//         ()  ,
//      
subjNameBLOB := pCertReqInfo^.Subject;
rsize := CertNameToStr(encType, @subjNameBLOB, CERT_SIMPLE_NAME_STR,
     subjNameString, 512);
subjectEdit.Text := subjNameString;

//      :
signCheckBox.Checked := CryptVerifyCertificateSignature(prov, encType,
                     reqEncoded, size,
                     @(pCertReqInfo^.SubjectPublicKeyInfo));
if not signCheckBox.Checked then
   begin
   case int64(GetLastError) of
   ERROR_FILE_NOT_FOUND: err := 'ERROR_FILE_NOT_FOUND';
   NTE_BAD_ALGID: err := 'NTE_BAD_ALGID';
   NTE_BAD_SIGNATURE: err := 'NTE_BAD_SIGNATURE';
   else err := 'CRYPT_E_OSS_ERROR';
   end;
   MessageDlg('Error verifying signature ' + err, mtError, [mbOK], 0);
   end;
//   -    " "  " "    ,
//     :
NotBeforeEdit.Text := DateTimeToStr(Now);
NotAfterEdit.Text := DateTimeToStr(Now + 365);

//      
StrDispose(subjNameString);
FreeMem(reqEncoded, size);
FreeMem(buf, 2048);
if not CryptReleaseContext(prov, 0) then
   MessageDlg('  ', mtError, [mbOK], 0);

end;


//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

procedure TSignCertForm.SignButtonClick(Sender: TObject);
var prov: HCRYPTPROV;
    cont: PChar;
    encType: DWORD;
    nameStr: PChar;
    nameAttr: CERT_RDN_ATTR;
    rdn: CERT_RDN;
    nameInfo: CERT_NAME_INFO;
    encNameLen: DWORD;
    encName: PBYTE;
    issNameBLOB: CERT_NAME_BLOB;
    err: string;
    certInfo: CERT_INFO;
    serNum: int64;
    params: CRYPT_OBJID_BLOB;
    sysTime: TSystemTime;
    encCertLen: DWORD;
    encCert: PByte;
    f: file;
begin
encType := PKCS_7_ASN_ENCODING or X509_ASN_ENCODING;
if length(ContainerEdit.Text) = 0
then cont := nil
else
    begin
    err := ContainerEdit.Text;
    cont := StrAlloc(length(err) + 1);
    StrPCopy(cont, err);
    end;

if not CryptAcquireContext(@prov, cont, MS_DEF_PROV, PROV_RSA_FULL, 0)
then
    begin
      MessageDlg('  : ' + err, mtError, [mbOK], 0);
      exit;
    end;

//    
certInfo.dwVersion := CERT_V3;
serNum := strtoint64(serNumEdit.Text);
certInfo.SerialNumber.cbData := sizeof(serNum);
certInfo.SerialNumber.pbData := @serNum;
FillChar(params, sizeof(params), 0);
certInfo.SignatureAlgorithm.pszObjId := szOID_OIWSEC_sha1RSASign;
certInfo.SignatureAlgorithm.Parameters := params;

nameStr := StrAlloc(length(IssuerEdit.text)+1);
StrPCopy(nameStr, IssuerEdit.Text);
nameAttr.pszObjId := '2.5.4.3';
nameAttr.dwValueType := CERT_RDN_PRINTABLE_STRING;
nameAttr.Value.cbData := length(IssuerEdit.text);
nameAttr.Value.pbData := PBYTE(nameStr);
rdn.cRDNAttr := 1;
rdn.rgRDNAttr := @nameAttr;
nameInfo.cRDN := 1;
nameInfo.rgRDN := @rdn;

//   ,    :
if not CryptEncodeObject(encType, X509_NAME,@nameInfo, nil, @encNameLen) or (encNameLen < 1) then
   begin
   MessageDlg('CryptEncodeObject error ' + inttostr(GetLastError), mtError, [mbOK], 0);
   StrDispose(nameStr);
   exit;
   end;

GetMem(encName, encNameLen);
if not CryptEncodeObject(encType, X509_NAME,@nameInfo, encName, @encNameLen) or (encNameLen < 1) then
   begin
   MessageDlg('2nd CryptEncodeObject error ' + inttostr(GetLastError), mtError, [mbOK], 0);
   StrDispose(nameStr);
   FreeMem(encName, encNameLen);
   exit;
   end;

issNameBlob.cbData := encNameLen;
issNameBlob.pbData := encName;
certInfo.Issuer := issNameBLOB;

//         :
DateTimeToSystemTime(StrToDateTime(NotBeforeEdit.Text), sysTime);
if not SystemTimeToFileTime(sysTime, certInfo.notBefore) then
   begin
   MessageDlg('Error converting system time to FILETIME', mtError, [mbOK], 0);
   StrDispose(nameStr);
   FreeMem(encName, encNameLen);
   exit;
   end;

DateTimeToSystemTime(StrToDateTime(NotAfterEdit.Text), sysTime);
if not SystemTimeToFileTime(sysTime, certInfo.notAfter) then
   begin
   MessageDlg('Error converting system time to FILETIME', mtError, [mbOK], 0);
   StrDispose(nameStr);
   FreeMem(encName, encNameLen);
   exit;
   end;

//   ,     ,    ,
//      :
certInfo.Subject := pCertReqInfo.Subject;
certInfo.SubjectPublicKeyInfo := pCertReqInfo.SubjectPublicKeyInfo;

//         :
certInfo.IssuerUniqueId.cbData := 0;
certInfo.IssuerUniqueId.pbData := nil;
certInfo.IssuerUniqueId.cUnusedBits := 0;
certInfo.SubjectUniqueId.cbData := 0;
certInfo.SubjectUniqueId.pbData := nil;
certInfo.SubjectUniqueId.cUnusedBits := 0;
certInfo.cExtension := 0;
certInfo.rgExtension := nil;

//     :
if not CryptSignAndEncodeCertificate(prov, AT_SIGNATURE, encType, X509_CERT_TO_BE_SIGNED,
   @certInfo, @(certInfo.SignatureAlgorithm), nil, nil, @encCertLen) then
   begin
     case int64(GetLastError) of
      ERROR_MORE_DATA: err := 'ERROR_MORE_DATA';
      ERROR_FILE_NOT_FOUND: err := 'ERROR_FILE_NOT_FOUND';
      NTE_BAD_ALGID: err := 'NTE_BAD_ALGID';
      else err := 'CRYPT_E_OSS_ERROR';
     end;
     MessageDlg('Error defining encoded certificate size: ' + err, mtError, [mbOK], 0);
     StrDispose(nameStr);
     FreeMem(encName, encNameLen);
     exit;
   end;

GetMem(encCert, encCertLen);

if not CryptSignAndEncodeCertificate(prov, AT_SIGNATURE, encType, X509_CERT_TO_BE_SIGNED,
   @certInfo, @(certInfo.SignatureAlgorithm), nil, encCert, @encCertLen) then
   begin
      MessageDlg('Error encoding certificate', mtError, [mbOK], 0);
      StrDispose(nameStr);
      FreeMem(encName, encNameLen);
      FreeMem(encCert, encCertLen);
      exit;
   end;

SaveDlg.Title := '  ?';
SaveDlg.Filter := ' (*.cer)|*.cer|  (*.*)|*.*';
SaveDlg.DefaultExt := 'cer';

if SaveDlg.Execute then
   begin
   AssignFile(f, SaveDlg.FileName);
   rewrite(f, 1);
   BlockWrite(f, encCert^, encCertLen);
   CloseFile(f);
   end;

//      
FreeMem(encCert, encCertLen);
FreeMem(encName, encNameLen);
StrDispose(nameStr);
if not CryptReleaseContext(prov, 0) then
   MessageDlg('  ', mtError, [mbOK], 0);

SignCertForm.Close;

end;

end.


